#ifndef file_hdr
#define file_hdr

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

class File;

#include "exc.h"

class Nop {};

/////////////////
// CLASS File //
///////////////
class File {

protected:
  char* file_name;
  int handle_d;
  int state_d;

public:
  File(const char* file_name, int flags,  int mode=S_IRUSR|S_IWUSR);
  ~File();

  void open(int flags=O_RDWR, int mode=S_IRUSR|S_IWUSR);
  void close();
  void read_file(char* buf, int size) const;
  void write_file(char* buf, int size) const;
  off_t seek_loc(off_t offset, int whence=SEEK_SET) const;
  off_t seek_end() const;
 
  const char* name() { return file_name; }
  int handle() { return handle_d; }
  int is_open() const { return handle_d >= 0; }

  friend ostream& operator<<(ostream&, const File&);

protected:
  File(const char* name, Nop):  
                   file_name(name?new char[strlen(name)+1]:(char*)NULL),
                   handle_d(-1), state_d(0)
  {
    UNEXPECTED( (!file_name),
	       "unexpected null string for file_name in File::File()" );
    strcpy(file_name,name);
  }

  virtual void output_data(ostream &)const;

private:
  File(const File&);
  void operator=(const File&);

public:
  static float read_time;
  static long times_read_called;
  static float write_time;
  static long times_write_called;
};

#endif
