/**********************************************************

	dbmisc.c:

***********************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <String.h>

#include "extern-c++.h"
#include "dbtemplate.h"
#include "psi_term.h"
#include "crypt.h"

extern "C" {
#include <pwd.h>
  char *getenv(const char *name);
  void check_psi_term(ptr_psi_term*);
  void check_definition(ptr_definition*);
}

#define STRNG(X) (char *)val[X]

static char *feature[] = {"1","2","3","4","5"};

// ptr_definition filestat;

/******** E_FILE_NAME(str, String)
  It returns the expenaded file_name in String, else return FALSE.
  For the time being all this does is replace '~' by the HOME directory
  if no user is given, or tries to find the user.

*/
static int e_file_name(char *s, String& es)
{
  char *home;
  struct passwd *pw;
  /* char *user="eight character name"; 18.5 */
  char userbuf[STRLEN];
  char *user=userbuf;
  char *t1,*t2;

  if (s[0]=='~') {
    t1=s+1;
    t2=user;
    while (*t1!=0 && *t1!='/') {
      *t2= *t1;
      *t2++;
      *t1++;
    }
    *t2=0;
    if ((int)strlen(user)>0) {
      pw = getpwnam(user);
      if (pw) {
	es += pw->pw_dir;
	es += t1;
	return TRUE;
      }
    }
    else {
      home=getenv("HOME");
      if (home) {
	es += home;
	es += (s+1);
	return TRUE;
      }
    }
  }  
  return FALSE;
}

static long c_expand_file()
{
  long success=FALSE;
  ptr_psi_term arg1,arg2,arg3,g;
  const char *file_name;
  String efile_name;

  g=aim->a;
  deref_ptr(g);
  get_two_args(g->attr_list,&arg1,&arg2);
  if(arg1) {
    deref(arg1);
    if (arg2) {
      file_name= PsiTerm:: psi_to_string(arg1);
      if (e_file_name(file_name, efile_name)) { file_name=(const char *)efile_name; }
      deref(arg2);
      deref_args(g,set_1_2);
      arg3=stack_psi_term(4);
      arg3->type=quoted_string;
      arg3->value=(GENERIC)heap_alloc(sizeof(file_name)); /* 12.5 */
      arg3->value=(GENERIC) heap_copy_string((char*)file_name);
      push_goal(unify,arg3,arg2,NULL);
    }
    success=TRUE;
  }
  else
    Errorline("no file name in %P.\n",g);

 return success;
}

static long stat_file(char* file_name, char* what, ptr_psi_term& arg) 
{
  long success = FALSE;
  ptr_psi_term t1;
  struct stat buf;
  long value;
  int f;
 
  if ( !stat(file_name,&buf) ) {
    arg=stack_psi_term(4);
    //  arg->type=filestat;
    
    switch (*what) {
    case '1':
      arg->type=integer;
      arg->value=(GENERIC)heap_alloc(sizeof(REAL)); /* 12.5 */
      value=(long)buf.st_mtime;
      *(REAL *)arg->value = (REAL) value;
      success=TRUE;
      break;
    case '2':
      arg->type=integer;
      arg->value=(GENERIC)heap_alloc(sizeof(REAL)); /* 12.5 */
      value=(long)buf.st_ino;
      *(REAL *)arg->value = (REAL) value;
      success=TRUE;
      break;
    default:
      success=TRUE;
    }
  }
  return success;
}

/******** C_FILE_STAT
  file_stat("file_name")
  Create a term into which information is placed concerning the file.
*/
static long c_file_stat()
{
  long success=FALSE;
  ptr_psi_term arg1,arg2,arg,g;
  const char *file_name;
  String efile_name;

  ptr_psi_term result = aim->b;
  deref(result);
  g=aim->a;
  deref_ptr(g);

  get_two_args(g->attr_list,&arg1,&arg2);
  if(arg1) {
    deref(arg1);
    if (arg2) {
      deref(arg2);
      deref_args(g,set_1_2);
      file_name = PsiTerm:: psi_to_string(arg1);
      if (e_file_name(file_name, efile_name)) { file_name=(const char*)efile_name; } 
      if (success=stat_file(file_name, arg2->type->keyword->symbol, arg)) {
	resid_aim=NULL;
	push_goal(unify,arg,result,NULL);
      }
    }
  }
  else
    Errorline("no file name in %P.\n",g);

  return success;
}

/******** C_ENCRYPT
  
  */
static long c_encrypt()
{
  long success=TRUE;
  ptr_psi_term funct,result,arg,arg1;
  char *val1;

  funct=aim->a;
  result=aim->b;
  deref_ptr(funct);
  deref_ptr(result);
  deref_args(funct, set_empty);

  get_one_arg(funct->attr_list,&arg1);

  if (arg1) {
    deref_ptr(arg1);
  }
  else {
    Errorline ("missing arguments in %P.\n", funct); 
    return FALSE;
  }

  if (arg1->value && sub_type(arg1->type,quoted_string)) 
    val1=(char *)arg1->value;
  else {
    Errorline("bad arguments in %P.\n", funct);
    return FALSE;
  }
  
  //  arg=stack_psi_term(0);
  arg=stack_psi_term(4);
  arg->type=quoted_string;
  arg->value=(GENERIC)heap_copy_string((const char*)Encode(val1));
  resid_aim=NULL;
  push_goal(unify,arg,result,NULL);
  
  return success;
}

/********************* C_IS_PERSISTENT_VAR ***********************/
/*	 C_IS_PERSISTENT_VAR -- LIFE func: is_persistent_var
	 is_persistent_var(+Var)
*/

static long c_is_persistent_var() {
  declare_var(1);
  
  begin_dbbuiltin(1,0);

  ptr_psi_term result = aim->b;
  deref(result);

  ptr_psi_term term = stack_psi_term(4);
  if (args[0]->type->type==global && 
      !((GENERIC)args[0]->type->global_value<(GENERIC)heap_pointer)) {
    term->type=true_type;
  }
  else term->type=false_type;
  resid_aim=NULL;
  push_goal(unify, term, result, NULL);

  end_dbbuiltin();
}

extern ptr_pair_list last_rule;
/******** C_RETRACT_ALL

*/
static long c_retract_all()
{
  declare_var(1);
  
  begin_dbbuiltin(1,0);

  if (args[0]->type->type==predicate || args[0]->type->type==function) {
    args[0]->type->rule=NULL;
    last_rule=NULL;
  }
  else {
    success=FALSE;
    Errorline("bad argument in retract_all.\n");
  }
  end_dbbuiltin();
}

/******** C_MEM_STAT

*/
static long c_mem_stat()
{
  ptr_psi_term t;
  long success=TRUE;
  long t1,t2,t3;

  t=aim->a;
  deref_args(t,set_empty);
  ptr_psi_term result = aim->b;
  deref(result);

  //  printf("Stack size  : %8d bytes (%5dK) (%ld%%)\n",t1,t1/1024,100*t1/t3);
  //  printf("Heap size   : %8d bytes (%5dK) (%ld%%)\n",t2,t2/1024,100*t2/t3);
  //  printf("Total memory: %8d bytes (%5dK)\n",t3,t3/1024);

  ptr_psi_term term = stack_psi_term(4);

  ptr_psi_term term1 = stack_psi_term(4);
  term1->type=integer;
  term1->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("stack_size",&(term->attr_list),term1);

  ptr_psi_term term2 = stack_psi_term(4);
  term2->type=integer;
  term2->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("heap_size",&(term->attr_list),term2);

  ptr_psi_term term3 = stack_psi_term(4);
  term3->type=integer;
  term3->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("memory",&(term->attr_list),term3);

  t1 = sizeof(mem_base)*(stack_pointer-mem_base);
  t2 = sizeof(mem_base)*(mem_limit-heap_pointer);
  t3 = sizeof(mem_base)*(mem_limit-mem_base);

  *(REAL *)term1->value =(REAL) t1;
  *(REAL *)term2->value =(REAL) t2;
  *(REAL *)term3->value =(REAL) t3;

  resid_aim=NULL;
  push_goal(unify, term, result, NULL);

  return success;
}

/******** C_SORTCMP

*/
static long c_sortcmp()
{
  ptr_psi_term result,funct,temp_result;
  ptr_psi_term arg1, arg2;
  long success=TRUE;

  long c_result;
  char *c_arg1=NULL;
  char *c_arg2=NULL; 
  REAL val1,val2;
  int is_val1, is_val2;

  funct=aim->a;
  deref_ptr(funct);
  result=aim->b;
  deref(result);

  get_two_args(funct->attr_list,&arg1,&arg2);
  if(arg1 && arg2) {
    deref(arg1);
    deref(arg2);
    deref_args(funct,set_1_2);
    if (arg1->value) {
      if (overlap_type(arg1->type,quoted_string)) {
	  c_arg1= (char *)arg1->value;
	  is_val1=1;
      }
      else if (overlap_type(arg1->type,integer)) {
	val1= *(REAL *)arg1->value;
	is_val1=1;
      }
      else {
	c_arg1=arg1->type->keyword->symbol;
	is_val1=0;
      }
    }
    else {
      c_arg1=arg1->type->keyword->symbol;
      is_val1=0;
    }

    if (arg2->value) {
      if (overlap_type(arg2->type,quoted_string)) {
	  c_arg2= (char *)arg2->value;
	  is_val2=1;
      }
      else if (overlap_type(arg2->type,integer)) {
	val2= *(REAL *)arg2->value;
	is_val2=1;
      }
      else {
	c_arg2=arg2->type->keyword->symbol;
	is_val2=0;
      }
    }
    else {
      c_arg2=arg2->type->keyword->symbol;
      is_val2=0;
    }

//     if (c_arg1) {
//       if (c_arg2) {
// 	c_result=strcmp( c_arg1, c_arg2 );
//       }
//       else {
// 	c_result=1;
//       }
//     }
//     else if (c_arg2) {
//       c_result=-1;
//     }
//     else {
//       c_result=(val1 < val2)? -1 :
// 	(val1==val2)? 0:1;
//     }
    //	resid_aim=NULL;
//     push_goal(unify,real_stack_psi_term(0,(REAL)c_result),result,NULL);

    if (is_val1) {
      if (is_val2) {
	// is_val1 and is_val2
	if (c_arg1) {
	  if (c_arg2) {
	    // both are strings
	    c_result=strcmp( c_arg1, c_arg2 );
	  }
	  else {
	    // first is string and second is real
	    c_result=1;
	  }
	}
	else if (c_arg2) {
	  // first is real and second is string
	  c_result=-1;
	}
	else {
	  // both are real
	  c_result=(val1 < val2)? -1 :
	    (val1==val2)? 0:1;
	}
      }
      else { 
	// is_val1 and not is_val2
	c_result=-1;
      }
    }
    else if (is_val2) { 
      // not is_val1 and is_val2
      c_result=1;
    }
    else { 
      // not is_val1 and not is_val2
      c_result=strcmp( c_arg1, c_arg2 );
    }
    ptr_psi_term res_t = stack_psi_term(4);
    res_t->type=integer;
    res_t->value=heap_alloc(sizeof(REAL));
    *(REAL *)res_t->value =(REAL) c_result;
    resid_aim=NULL;
    push_goal(unify, res_t, result, NULL);

  }
  else {
    Errorline("argument missing in %P.\n",funct);
    success=FALSE;
  }

  return success;
}

void check_misc_definitions() {
  //  check_definition(&filestat);
}

void insert_misc_builtins()
{
  ptr_module curmod = current_module;
  set_current_module(db_module);

  //  filestat = update_symbol(db_module,"filestat");
  new_built_in(db_module,"file_stat",function,c_file_stat);
  new_built_in(db_module,"expand_file",predicate,c_expand_file);
  new_built_in(db_module, "is_persistent_var", function, 
	       c_is_persistent_var);
  new_built_in(db_module,"encrypt",function,c_encrypt);
  new_built_in(db_module,"mem_stat",function,c_mem_stat);
  new_built_in(db_module,"sortcmp",function,c_sortcmp);
  new_built_in(db_module, "retract_all", predicate, c_retract_all);
  set_current_module(curmod);
}

