#ifndef o_buffer_hdr
#define o_buffer_hdr
#define NO_OBJECT_PAGE

#ifndef NO_OBJECT_PAGE
class ObjectPageNode;
#endif
class ObjectBuffer;
class ObjectNode;

#include "oid.h"
#include "dlist.h"
#include "slist.h"
#include "hash_mat3.h"
#include "hash_mat2.h"
#include "psi_term.h"
  
//////////////////////
// ObjectPageNode //
////////////////////
#ifndef NO_OBJECT_PAGE
class ObjectPageNode:public B_DL_Node {

private:
  const PageKey key_d;
  FList data_d; // since FList is on LIFE heap, make sure the first and last 
               // next,prev pointers are garbage collected.
  static int n_of_nodes_d;
public:
  ObjectPageNode(const PageKey& p_key):key_d(p_key)
  {
    UNEXPECTED( (this != (void*)BYTEDATA_DATA(alloc_ptr)),
	       "unexpected error in ObjectPageNode:Id class" )
  }
  ~ObjectPageNode();

  //  int modified_object(ObjectNode*& hn, ObjectPage* & page);
  void flush_object_page(); // object_page flushed if file writable.
  const PageKey& key() const {return key_d;}
  FList& data() const { return data_d; }


  void* operator new(size_t s) {
    alloc_ptr = PsiTerm::byte_data_alloc(s);
    n_of_nodes_d++;
    return (void*)BYTEDATA_DATA(alloc_ptr);
  }
  void operator delete(void *) { n_of_nodes_d--; }
  static int num_of_nodes() { return n_of_nodes_d; }

  void check_data();

protected:
  ObjectNode* next_object(ObjectNode* obn);
  void output_data(ostream&)const;


private:
  static GENERIC alloc_ptr;
};
#endif

////////////////////////
// CLASS ObjectNode //
//////////////////////
#ifndef NO_OBJECT_PAGE
class ObjectNode:public B_SL_Node {
  friend void ObjectPageNode::flush_object_page();
#else
class ObjectNode:public B_FL_Node {
#endif
  friend class ObjectBuffer;

private:
  GENERIC key_d;		// functions as key to the node
  ptr_psi_term term_d;
  static GENERIC alloc_ptr;
  static int n_of_nodes_d;

class Id: public ObjectId {
public:
  Id(const ObjectKey& o_key):ObjectId(o_key, alloc_ptr)
  {
    UNEXPECTED( (this != (void*)BYTEDATA_DATA(alloc_ptr)),
	       "unexpected error in ObjectNode:Id class" )
  }

  void* operator new(size_t s) {
    alloc_ptr = PsiTerm::byte_data_alloc(s);
    return (void*)BYTEDATA_DATA(alloc_ptr);
  }

  void operator delete(void *) { }

private:
  static GENERIC alloc_ptr;

};

public:
  ObjectNode(const ObjectKey& o_key):term_d(heap_psi_term(4)) 
{
  UNEXPECTED( (this != (void*)BYTEDATA_DATA(alloc_ptr)),
	      "unexpected error in ObjectNode class" );

  ObjectId* object_id = new Id(o_key);
  key_d = object_id->dbpointer();
  term_d->type=psi_id;
  term_d->value=NULL;
  term_d->dbpointer = key_d;
  key().hashed_object();
}
  
  ObjectNode(const ObjectKey& o_key, ptr_psi_term term):
                                 term_d(heap_psi_id_term(term))
{
  deref_ptr(term_d);
  ObjectId* object_id = new Id(o_key);
  key_d = object_id->dbpointer();
  term_d->dbpointer = key_d;
  key().hashed_object();
}

  ObjectNode(ptr_psi_term term): term_d(term)
{
  deref_ptr(term_d);
  UNEXPECTED( ( (GENERIC)term_d < (GENERIC)heap_pointer ),
	      "unexpected stack pointer in ObjectNode::ObjectNode()");

  key_d = term_d->dbpointer;
  key().dbcoreffed_psi_term(term_d->dbpointer);
  key().hashed_object();
}
  ~ObjectNode();
  void check_data();
  int is_modified();

  ObjectId& key() const {return PSIID(key_d);}
  ptr_psi_term term() { deref_ptr(term_d); return term_d; }
  GENERIC dbpointer() { return key_d; }  
  void flush_object();

  void* operator new(size_t s) {
    alloc_ptr = PsiTerm::byte_data_alloc(s);
    n_of_nodes_d++;
    return (void*)BYTEDATA_DATA(alloc_ptr);
  }
  void operator delete(void *) {n_of_nodes_d--;}
  static int num_of_nodes() { return n_of_nodes_d; }

  friend ostream& operator<<(ostream& os, const ObjectNode& hn);

private:
  ptr_psi_term heap_psi_id_term(ptr_psi_term&);
  void output_data(ostream&)const;

};


typedef HashMatrix3<ObjectKey, ObjectNode> ObjHashTable;
#ifndef NO_OBJECT_PAGE
typedef HashMatrix2<PageKey, ObjectPageNode> PageObjsTable;
#endif

/////////////////////////
// CLASS ObjectBuffer //
///////////////////////
#ifndef NO_OBJECT_PAGE
class ObjectBuffer:public SList {
#else
class ObjectBuffer:public FList {
#endif

private:
  ObjHashTable* obj_table;
	  // do not delete obj_table as the modified hashed_psi_terms 
         // need to be flushed to the file. 
#ifndef NO_OBJECT_PAGE
  PageObjsTable* page_objs_table;
#endif

public:
  ObjectBuffer(int object_hash_size=7, 
	       int page_hash_size=13, 
	       int file_hash_size=5);

  int flush_object_buffer(int save_all);
  int is_flushed();
  int evict_objects(int min_num=0);
  int evict_objects(ObjFile*);

  REAL size() { return ObjectNode::num_of_nodes()*sizeof(ObjectNode); }
  int length() { return ObjectNode::num_of_nodes(); }
  void check_data();

  void usage(ostream& os);
  void output_data(ostream&)const;

public:
  ptr_psi_term get_psi_term(const ObjectKey& o_key);
  ptr_psi_term get_psi_term(const ObjectKey& o_key, GENERIC&);
  ptr_psi_term insert_psi_term(const ObjectKey& o_key);
  ptr_psi_term insert_psi_term(const ObjectKey& o_key, ptr_psi_term term);
  ptr_psi_term insert_psi_term(ptr_psi_term term);

private:
  ObjectNode* first_object() {
    return (ObjectNode*) first();
  }
  ObjectNode* next_object(ObjectNode* obn) {
#ifndef NO_OBJECT_PAGE
    return (ObjectNode*)obn->next3();
#else
    return (ObjectNode*)obn->next2();
#endif
  }
  void mv2end(ObjectNode* obn) {
#ifndef NO_OBJECT_PAGE
    obn->remove3();
#else
    obn->remove2();
#endif
    //    cerr << "ObjectNode::size: " << ObjectNode::num_of_nodes() << "\n";
    append(obn);
  }
  void flush_object(ObjectNode* obn) {
#ifndef NO_OBJECT_PAGE
    ObjectPageNode* object_page = get_object_page(obn->key());
    object_page->flush_object_page();
#else
    obn->flush_object();
#endif
  }

#ifndef NO_OBJECT_PAGE
  void init_object_page_iterator();
  ObjectPageNode* next_object_page();
  ObjectPageNode*  get_object_page(const PageKey& p_key);
         // get_object_page return a list of objects which
	 // are in the same page.
  ObjectPageNode*  make_object_page(const PageKey& p_key);
  void append2object_page(ObjectNode* obn) {
    ObjectKey& o_key = obn->key();
    ObjectPageNode* object_page = get_object_page(o_key);
    if (!object_page) object_page = make_object_page(o_key);
    object_page->data().append(obn); 
  }
#endif
  //  int buf_length();
};

#endif
