#include <fcntl.h>

#include "obj_file.h"
#include "exc.h"
#include "obj_page.h"
#include "oid.h"
#include "p_buffer.h"
#include "o_buffer.h"
#include "psi_obj.h"
#include "defs.h"

extern "C" {
  // #include "Source/memory.h"
  // #include "Source/trees.h"
  void check_psi_term(ptr_psi_term*);
  int system(const char *string);
  
};

///////////////////
// CLASS DbFile //
/////////////////
DbFile:: DbFile(const char* name, DbFileInfo* info, Nop): File(name, Nop()),
                   info_d(info)
{
}

DbFile:: ~DbFile() {
}

void DbFile:: output_data(ostream& os)const{
  File::output_data(os);
  os << ", "  << *info_d;
}

////////////////////
// CLASS ObjFile //
//////////////////

ObjFile:: ObjFile(const char* name): 
                      DbFile(name, new ObjFileInfo, Nop()) ,
		      file_writable_d(files_writable)
{

  ObjFileInfo& info = *(ObjFileInfo*)info_d;

  if (file_writable_d)
    handle_d = ::open(file_name, O_RDWR);
  else
    handle_d = ::open(file_name, O_RDONLY);
  if ( handle_d == -1 ) {
    //#include <iostream.h>
    if (errno == ENOENT && file_writable_d) {	
                                 // file does not exist, create it
				// with a default page header
      {
	handle_d = ::open(file_name, O_RDWR | O_CREAT, S_IRUSR|S_IWUSR);
	if ( handle_d == -1 ) {
	  dberror << "could not open file -- \"" 
	    << file_name << "\" for output " ;
	  throw_error();
	}
	info.init();
	info.num_of_obj_pages = info.last_obj_page = 1;
	PageKey p_key(this, info.last_obj_page);
	ObjectPage p(p_key, FirstNewPage());
	info.modified(1);
	//	flush_file();
	flush_file_hdr();
	p.flush_page();
      }
    }
    else {
      if (!file_writable_d) 
	dberror << "could not open file -- \"" << file_name 
		<< "\" for reading";
      else
	dberror << "could not open file -- \"" << file_name 
		<< "\" for writing";
      throw_error();
    }
  }
  else {      
    seek_loc(0);
    char buf[ObjFileInfo::FILE_HDR_SIZE];
    read_file(buf, ObjFileInfo::FILE_HDR_SIZE);
    info.read_hdr(buf);
  }
}

ObjFile:: ~ObjFile() {
}

// void ObjFile:: flush_file() {
//   if (file_writable()) {
//     PageObjectsNode* pon = (PageObjectsNode*) file_objects->first();
//     print_debug((cerr << "*** Flushing objects of " << name() << "\n"));
//     while (pon) {
//       pon->flush_object_page();
//       pon = (PageObjectsNode*)pon->next2();
//     }
//     PageNode* pn = (PageNode*) file_pages->first();
//     print_debug((cerr << "*** Flushing pages of " << name() << "\n"));
//     while (pn) {
//       pn->flush_page();
//       pn = (PageNode*)pn->next2();
//     }
//     flush_file_hdr();
//   }
// }

void ObjFile:: flush_file_hdr() {
  if (file_writable()) {
    ObjFileInfo& info = *(ObjFileInfo*)info_d;
    if (info.is_modified()) {
      char buf[ObjFileInfo::FILE_HDR_SIZE];  
      print_debug((cerr << "*** Flushing file header of " << name() << "\n"));
      seek_loc(0);
      info.write_hdr(buf);
      write_file(buf, ObjFileInfo::FILE_HDR_SIZE);
      info.modified(0);
    }
  }
}

page_num ObjFile:: new_page_num(const page_num& prev_page_num) {
//   ObjFileInfo& info = *(ObjFileInfo*)info_d;
//   ++info.num_of_obj_pages;
//   if (info.last_obj_page == prev_page_num)
//     info.last_obj_page = info.num_of_obj_pages;
//   info.modified(1);
//   return info.num_of_obj_pages;

  ObjFileInfo& info = *(ObjFileInfo*)info_d;
  ++info.num_of_obj_pages;
  info.modified(1);
  return info.num_of_obj_pages;
}

off_t ObjFile:: seek_page(const page_num& pnum) const {
  return seek_loc(ObjFileInfo::pnum2bid(pnum));
}

void ObjFile:: output_page(ostream& os, page_num pnum){
  const char* data;
  PageKey p_key(this, pnum);
  ObjectPage page(p_key);
  int num_of_objects=page.num_of_objects();
  for (int i=page.first_object(); i<=num_of_objects; i++) {
    data = page[i];
//    Object2Psi obj (this, data);
    os << i << ") ";
//    os << obj;
    os << "\n";
  }
}

void ObjFile:: output_data(ostream& os){
  DbFile::output_data(os);
  ObjFileInfo& info = *(ObjFileInfo*)info_d;
  page_num last=info.last_obj_page;
  os << "\n\n";
  for (page_num i=info.first_obj_page; i <= last; i++) {
    os << "*** Object Page " << i << "*******\n";
    output_page(os, i);
  }
}


void ObjFile:: usage(ostream& os) {
  os << "\"" << name() <<
    "\" => number of pages " << num_of_obj_pages() << "\n";
  page_num last_page = last_obj_page();
  for(int i=first_obj_page(); i <= num_of_obj_pages(); i++) {
    ObjectPage page(PageKey(this,i));
    os << "*** Page: "; 
    page.usage(os); 
    os << "\n";
    page.page_ok();
  } 
}

// void ObjFile:: insert_dbpers_var(ptr_psi_term t) {

//   deref_ptr(t);
//   UNEXPECTED( ( (GENERIC)t < (GENERIC)heap_pointer ),
// 	      "unexpected stack pointer in ObjFile:: insert_dbpers_var");

//   if (!t->dbpointer) {
//     DBPersVarNode* pvn = new DBPersVarNode(t);
//     dbpers_vars_d.append(pvn);
//   }

// }

// void ObjFile:: check_dbpers_vars() {
//   DBPersVarNode* pvn = (DBPersVarNode*) dbpers_vars_d.first();
//   while(pvn) {
//     check_psi_term(&(pvn->t));
//     pvn = (DBPersVarNode*) pvn->next1();
//   }
// }

// int ObjFile:: flush_dbpers_vars(int save_all) {
//   int flushed=1;
//   if (file_writable) {
//     DBPersVarNode* pvn = (DBPersVarNode*) dbpers_vars_d.first();
//     DBPersVarNode* n_pvn = NULL;
//     ptr_psi_term arg;
//     while(pvn) {
//       arg = pvn->t;
//       deref_ptr(arg);
//       if (!arg->dbpointer) PsiObject obj(arg, this);
//       n_pvn = (DBPersVarNode*) pvn->next1();
//       delete pvn;
//       pvn=n_pvn;
//     }
//   }
//   return flushed;
// }

