#ifndef obj_stream_hdr
#define obj_stream_hdr

class ObjFile;
class PageBuffer;
class ObjectBuffer;
class Buffer;
class PageKey;

#include "oid.h"
#include "dlist.h"
#include "hash.h"

class FileNode: public B_DL_Node {
private:
  ObjFile* file_d;

public:
  FileNode(ObjFile* f):file_d(f) {}
  ~FileNode(); 

  ObjFile* file() { return file_d; }

protected:
  void output_data(class ostream &) const {}
};

class ObjStream {

private:
  DList file_table;
  // Care should be taken that the files are not opened twice, although
  // on LIFE side it may be better to check the file disk i-node value here
  // to see that a file is not opened twice.

  //  ObjFile* curr_obj_file;

  PageBuffer* page_buffer_d;
    // page_buffer_d(int max_page_bufs, int page_hash_size=13,  
    //          int file_hash_size=5)

  ObjectBuffer* object_buffer_d;
       // object_buffer_d(int object_hash_size=7, 
       //                 int page_hash_size=13, int file_hash_size=5)

public:

  //  ObjStream(int max_pages=10, int obj_y_hash_size=10);
  ObjStream(REAL max_pages=10.0, REAL obj_y_hash_size=10.0);
  ~ObjStream(); // ObjStream destructor flushes all the files.

  //  ObjFile* obj_file() { return curr_obj_file; }
  ObjFile* open_file(char* str);
  int close_file(ObjFile*);
  //  void set_file(ObjFile* f) {curr_obj_file=f;}
  //  void close_file() {}

  PageBuffer* page_buffer() { return page_buffer_d; }
  ObjectBuffer* object_buffer() { return object_buffer_d; }

  //  void flush_obj_stream();
  int flush_obj_stream(int save_all=0);

  void check_data();
  int evict_objects(int min_pages=0, int min_objects=0);
  void output_heap_info(ostream&);
  void output_statistics(ostream&);
  friend ostream& operator<<(ostream& os, const ObjStream& n);

protected:
  virtual void output_data(ostream& os)const;

private:
  static int page_hash_table_size[2];
         // [0] = file_hash_size, [1] = page_hash_size
  static int obj_hash_table_size[3];
	// [0] = file_hash_size, [1] = page_hash_size, [3] = object_hash_size

public:
  void usage(ostream& os)const;
};

class Odb {
private:
  ObjStream* o_stream;
public:
  Odb():o_stream(NULL) {}
  ~Odb() { delete o_stream; }
  void operator()(ObjStream* o) {o_stream=o;}
  ObjStream* operator->() {return o_stream;}
  ObjStream& operator*() {return *o_stream;}
  int is_set() { return o_stream?1:0; }
};

#endif
