#include "exc.h"
#include "odb.h"

//////////////////////
// STRUCT PageInfo //
////////////////////
const u_1byte PageInfo:: type_d=PAGE;

PageInfo:: PageInfo(){
}

int PageInfo::  read_hdr(char* buf) {
  if (type() != PAGE_TYPE(buf)) return 0;
  return 1;
}

void PageInfo:: write_hdr(char* buf)const{
  PAGE_TYPE(buf) = type();
}

ostream& operator<<(ostream& os, const PageInfo& info) {
  info.output_data(os);
  return os;
}

/////////////////////////
// STRUCT ObjPageInfo //
///////////////////////
const u_1byte ObjPageInfo:: type_d=OBJ_PAGE;

ObjPageInfo:: ObjPageInfo() {
}

void ObjPageInfo:: init(char* buf) {
  memset(buf, ' ', PAGE_SIZE);
  next_page = 0;
  prev_page = 0;
  free_slot = 1;
  free_loc = FIRST_OBJ_LOC_IN_OBJ_PAGE;
  //  free_bytes = OBJ_SIZE_MAX;
  free_bytes = PAGE_SIZE-OBJ_PAGE_HDR_SIZE;
  actual_free_bytes = free_bytes;
}

int ObjPageInfo:: read_hdr(char* buf) {
  int succ=1;
  succ=PageInfo:: read_hdr(buf);
  pt2p_4byte(next_page, &buf[1]);
  pt2p_4byte(prev_page, &buf[5]);
  pt2s_2byte(free_slot, &buf[9]);
  pt2s_2byte(free_loc, &buf[11]);
  pt2s_2byte(free_bytes, &buf[13]);
  pt2s_2byte(actual_free_bytes, &buf[15]);
  return succ;
}

void ObjPageInfo:: write_hdr(char* buf)const{
  PageInfo::write_hdr(buf);
  cp_p_4byte_to_pt(&buf[1], next_page);
  cp_p_4byte_to_pt(&buf[5], prev_page);
  cp_s_2byte_to_pt(&buf[9], free_slot);
  cp_s_2byte_to_pt(&buf[11], free_loc);
  cp_s_2byte_to_pt(&buf[13], free_bytes);
  cp_s_2byte_to_pt(&buf[15], actual_free_bytes);
}

void ObjPageInfo:: usage(ostream& os)const{
  os << "space left: " << free_bytes << "(";
  os << "actual: " << actual_free_bytes << ")";
}

void ObjPageInfo:: output_data(ostream& os)const{
  //  os << "Type: OBJ_PAGE" << "\n";
  //  os << "next_page: " << next_page << " ";
  //  os << "prev_page: " << prev_page << " ";
  os << "free_slot: " << free_slot << " ";
  os << "free_loc: " << free_loc << " ";
  os << "free_bytes: " << free_bytes << " ";
  os << "actual_free_bytes: " << actual_free_bytes;
}

////////////////////////
// STRUCT DbFileInfo //
//////////////////////
const u_1byte DbFileInfo:: type_d=DB_FILE;

ostream& operator<<(ostream& os, const DbFileInfo& info) {
  info.output_data(os);
  return os;
}

////////////////////////
// STRUCT ObjFileInfo //
///////////////////////
//const TwoByteLog ObjFileInfo:: FILE_HDR_SIZE(10, LogType());
const u_2byte ObjFileInfo:: FILE_HDR_SIZE=1024;
   // byte 0,1,2,3 = last_obj_page in the file
   // byte 4,5,6,7 = num_of_obj_pages in the file.

const u_1byte ObjFileInfo:: type_d=OBJ_FILE;

const page_num ObjFileInfo:: first_obj_page = 1; 

ObjFileInfo:: ObjFileInfo():last_obj_page(0), num_of_obj_pages(0) {
  // page_num of 0 is used to indicate NULL value
}

page_num ObjFileInfo:: bid2pnum(const p_4byte& bid){
  return ((bid-FILE_HDR_SIZE)/PAGE_SIZE)+1;
}

p_4byte ObjFileInfo:: pnum2bid(const page_num& pnum){
//  return ( ((pnum-1)*PAGE_SIZE) + FILE_HDR_SIZE );
  return ( FILE_HDR_SIZE + ((pnum-1)*PAGE_SIZE) );
}

void ObjFileInfo:: init() {
  last_obj_page = 0;
  num_of_obj_pages = 0;
  state_d = 0;
}

void ObjFileInfo:: write_hdr(char* buf)const{
  memset(buf,' ',FILE_HDR_SIZE);
  cp_p_4byte_to_pt(buf, last_obj_page);
  cp_p_4byte_to_pt(&buf[4], num_of_obj_pages);
}

void ObjFileInfo:: read_hdr(const char* buf) { 
  pt2p_4byte(last_obj_page, buf);
  pt2p_4byte(num_of_obj_pages, &buf[4]);
}
 
void ObjFileInfo:: output_data(ostream& os)const{
//  os << "Type: object file";
//   os << "( ";
//   os << "Num of Pages: " << num_of_obj_pages << ", ";
//   os << "First Page: " << first_obj_page << ", ";  
//   if (last_obj_page) {
//     os << "Last Page: " << last_obj_page;
//   }
//   else {
//     os << "Last Page: NULL";
//   }
//   os << " )";
  os << "Number of Pages: " << num_of_obj_pages;
}

