/**********************************************************

	db_builtins.cpp:
	Database functions and predicates for
	the wild_LIFE interpreter v1.02

***********************************************************/

#include "extern-c++.h"
#include "dbtemplate.h"
#include "psi_term.h"
#include "obj_file.h"
#include "page.h"
#include "p_buffer.h"
#include "oid.h"
#include "psi_obj.h"
#include "obj_stream.h"
#include "o_buffer.h"

extern "C" {

void init_db_builtins();
void check_db_definitions();
void check_psi_term(ptr_psi_term*);
void check_definition(ptr_definition*);
int c_main(long, char**);
//ptr_node copy_attr_list(ptr_node);
#include "lock.h"
}

extern void insert_misc_builtins();
extern void check_rdb_definitions();
extern void check_misc_definitions();
extern void init_rdb_builtins();

#define DBCHANNEL(X) *(ObjFile**)BYTEDATA_DATA(X)

#define STRNG(X) (char *)val[X]
#define INT(X) (long)*(REAL *)val[X]
#define REAL_VAL(X) *(REAL *)val[X]

/****************Externally defined variables*********/
extern ptr_definition true_type;
extern ptr_definition false_type;

/****************Global Varaibles********************/
ptr_definition odb_file;
ptr_definition buffer_group;
ptr_definition psi_id; 
// if term->type==psi_id then term->value should be NULL.
ptr_definition dbstore_handlersym;

//ObjStream* object_stream=NULL;
Odb object_stream;
int files_writable=0;
int save_top_only = 1; 
        // only the top is saved, and the object is marked as 
       // unsaved. if this value changed to 0 in any part of code
      // it should be changed back to 1 before calling c_odb_store_psi.
int fetch_depth=0;

static char *feature[] = {"1","2","3","4","5"};

/****************Externally defined functions*********/

/****************Forward Declarations*****************/


/********************* C_ODB_INIT ***********************/
/*	 C_ODB_INIT- LIFE pred: odb_init
	 odb_init(+X)
	 X == max (% of alloc_words) pages that can be allocated from
	 the LIFE heap
	 Y == size (% of alloc_words/sizeof(psiterm+objectnode)) of
         y index of object hash table
*/
static long c_odb_init() {
  ptr_psi_term psiTerm, arg1, arg2;
  long success=FALSE;

   psiTerm=aim->a; 
   deref_ptr(psiTerm); 
   get_two_args(psiTerm->attr_list,&(arg1),&(arg2));
   if (arg1 && arg2) {
     deref(arg1);
     deref(arg2);
     deref_args(psiTerm,set_1_2);
     if (arg1->value && arg2->value
 	&& sub_type(arg1->type, real) 
 	&& sub_type(arg2->type, real)){
       success=TRUE;
     }
   }
   if (success) {
     if (object_stream.is_set()==0) 
       object_stream(new ObjStream(*(REAL*)arg1->value,
				   *(REAL*)arg2->value));
   }
   else 
    Errorline ("bad arguments in %P.\n", psiTerm); 
  return success;

  //  if (object_stream.is_set()==0) 
  //       object_stream(new ObjStream(2,1));
   
  //  return TRUE;
}

/********************* C_ODB_STORE_PSI ***********************/
/*	 C_ODB_STORE_PSI-- LIFE pred: odb_store_psi
	 odb_store_psi(+psi_term)
	 odb_store_psi(+psi_term, +odb_file)
*/
static long c_odb_store_psi() {
  declare_var(1);

  begin_dbbuiltin(1,0);

  ptr_psi_term term;

  if ((GENERIC)args[0] < heap_pointer) {
    //    Errorline("%P is not a persistent term\n",args[0]);
    //    success=FALSE;
    term=heap_psi_term(4);
    push_psi_ptr_value(args[0], &(args[0]->coref));
    args[0]->coref=term;
  }
  else {
    term=args[0];
    if (term->dbpointer) {
      PsiObject obj(term);
    }
    else {
      ObjFile* file;
      ptr_psi_term arg2 = PsiTerm::get_arg(psiTerm, "2");
      if (arg2) {
	deref_ptr(arg2);
	
	if ( (success = (arg2->type == odb_file) && (arg2->value != NULL)) ){
	  file =  DBCHANNEL(arg2->value);
	  PsiObject obj(term, file);
	}
	else 
	  Errorline ("bad arguments in %P.\n", psiTerm);
      }
      else {
	Errorline("missing arguments in %P.\n", psiTerm);
	success=FALSE;
      }
    }
  }
  end_dbbuiltin();

}

/********************* ODB_OPEN_FILE ***********************/
/*	C_ODB_OPEN_FILE -- LIFE pred: odb_open_file
	odb_open_file(+FileName, -DbChannel)
*/
static long c_odb_open_file()
{
  declare_var(2);

  types[0] = quoted_string;

  begin_dbbuiltin(2,1);

  if (is_top(args[1])) {

    ptr_psi_term psi_db_channel = 
      PsiTerm::heap_bd_psi_term( args[1], odb_file, 
				 sizeof(ObjFile*) );
    deref_ptr(psi_db_channel);
    GENERIC bd = psi_db_channel->value;
    DBCHANNEL(bd) = object_stream->open_file(STRNG(0));
  }
  else {
    Errorline("bad argument in %P.\n",psiTerm);
    success=FALSE;
  }

  end_dbbuiltin();
  
}

/********************* C_ODB_GET_FILE ***********************/
/*	 C_ODB_GET_FILE-- LIFE pred: odb_get_file
	 odb_get_file(+odb_file)
*/
static long c_odb_get_file() {
  declare_var(1);

  types[0] = odb_file;
  ptr_psi_term result=aim->b;
  deref(result);

  begin_dbbuiltin(1,1);
  
  ObjFile* file = DBCHANNEL(val[0]);
  //  ptr_psi_term arg1=stack_psi_term(0);
  ptr_psi_term arg1=stack_psi_term(4);
  arg1->type=quoted_string;
  arg1->value=(GENERIC)heap_copy_string(file->name());
  resid_aim=NULL;
  push_goal(unify,arg1,result,NULL);
  
  end_dbbuiltin();
}

/********************* C_ODB_SET_FILE_WRITABLE ***********************/
/*	 C_ODB_SET_FILE_WRITABLE-- LIFE pred: odb_set_file_writable
	 odb_set_file_writable(+int)
	 int=1: if odb_open_file("F") has been called with files_writable=1,
                then whenever the persistent psi_terms for this file
                is updated they are flushed to the file, otherwise the
                file is untouched
*/
static long c_odb_set_file_writable() {
  declare_var(1);

  types[0]=integer;

  begin_dbbuiltin(1,1);
  
  files_writable = (INT(0) > 0)?1:0;

  end_dbbuiltin();

}

/********************* C_ODB_GET_STREAM ***********************/
/*	 C_ODB_GET_STREAM-- LIFE pred: odb_get_stream
	 odb_get_stream(+psi_term)
*/
static long c_odb_get_stream() {
  declare_var(1);
  begin_dbbuiltin(1,0);
  ptr_psi_term result = aim->b;
  deref(result);

  if (args[0]->dbpointer) {
    ptr_psi_term term = heap_psi_term(4);
    term = PsiTerm::heap_bd_psi_term( term, odb_file, 
				      sizeof(ObjFile*) );
    deref_ptr(term);
    GENERIC bd = term->value;
    ObjectId& o_id = PSIID(args[0]->dbpointer);
    DBCHANNEL(bd) = o_id.file();
    resid_aim=NULL;
    push_goal(unify,term,result,NULL);
  }
  else {
    Errorline("bad argument in %P.\n",psiTerm);
    success=FALSE;
  }

  end_dbbuiltin();

}

/********************* C_ODB_FLUSH_STREAM ***********************/
/* C_ODB_FLUSH_STREAM-- LIFE pred: odb_flush_stream
 odb_flush_stream()
*/
static long c_odb_flush_stream() {
 
  declare_var(0);
  
  begin_dbbuiltin(0,0);
  // ptr_psi_term result = aim->b;
  // deref(result);
  //  save_top_only=0;
  if (object_stream.is_set())
    success=object_stream->flush_obj_stream();
  //  save_top_only=1;
  // resid_aim=NULL;
  // push_goal(unify, id, result, NULL);
  end_dbbuiltin();

}

/********************* C_ODB_CLOSE_STREAM ***********************/
/*	 C_ODB_CLOSE_STREAM-- LIFE pred: odb_close_stream
	 odb_close_stream(+odb_file)
*/
static long c_odb_close_stream() {
  declare_var(1);

  types[0] = odb_file;

  begin_dbbuiltin(1,1);
  
  ObjFile* file = DBCHANNEL(val[0]);
  success = (object_stream->close_file(file));
  end_dbbuiltin();
}

void call_dbstore_handler(ptr_psi_term term) {

  ptr_psi_term dbstore;
  
  dbstore=stack_psi_term(0);
  dbstore->type=dbstore_handlersym;
  stack_add_psi_attr(dbstore,"1", term); 
  push_goal(prove,dbstore,DEFRULES,NULL);
  
}

/********************* C_ODB_TERM2ID ***********************/
/*	 C_ODB_TERM2ID -- LIFE func: odb_term2id
	 odb_term2id(+pid, +sid)
	 odb_term2id(+pid, +sid, +odb_file)
*/

static long c_odb_term2id() {
  declare_var(2);
  
  types[0]=integer;
  types[1]=integer;

  begin_dbbuiltin(2,2);

  ptr_psi_term result = aim->b;
  deref(result);

  ObjFile* file;
  ptr_psi_term arg3 = PsiTerm::get_arg(psiTerm, "3");
  if (arg3) {
    deref_ptr(arg3);    
    if ( (success = (arg3->type == odb_file) && (arg3->value != NULL)) )
      file =  DBCHANNEL(arg3->value);
    else
      Errorline ("bad arguments in %P.\n", psiTerm);
  }
  else {
    Errorline("missing argument in %P.\n", psiTerm);
    success=FALSE;
  }
    
  if (success) {
    ptr_psi_term id = PsiTerm::psi_term2psi_id(args[0], args[1], file);
    if (id) {
      deref_ptr(id);
      ObjectId& object_id = PSIID(id->dbpointer);
      //      object_id.pers_var();
      if (object_id.pnum() > file->num_of_obj_pages()) {
	Errorline("invalid database reference in %P for file %s \n",
		  psiTerm, file->name());
	success=FALSE;
      }
    resid_aim=NULL;
    push_goal(unify, id, result, NULL);
    }
    else {
      success = FALSE;
    }
  }
  end_dbbuiltin();
}

/********************* C_ODB_ID2TERM ***********************/
/*	 C_ODB_ID2TERM -- LIFE func: odb_id2term
	 odb_id2term(+psi_id)
*/

static long c_odb_id2term() {
  declare_var(1);
  
  begin_dbbuiltin(1,0);

  ptr_psi_term result = aim->b;
  deref(result);
  if (args[0]->dbpointer) {
    ptr_psi_term term = PsiTerm:: psi_id2psi_term(args[0]);
    resid_aim=NULL;
    push_goal(unify, term, result, NULL);
  }
  else {
    Errorline("bad argument in %P.\n",psiTerm);
    success = FALSE;
  }

  end_dbbuiltin();
}

/********************* C_ODB_FETCH_DEPTH ***********************/
/*	 C_ODB_FETCH_DEPTH- LIFE pred: odb_fetch_depth
	 odb_fetch_depth(+X)
*/
static long c_odb_fetch_depth() {
  declare_var(1);
  types[0]=integer;

  begin_dbbuiltin(1,1);

  fetch_depth = INT(0);
  cerr << "fetch_depth: " << fetch_depth << "\n";
  end_dbbuiltin();

}

/********************* C_ODB_IS_PERSISTENT ***********************/
/*	 C_ODB_IS_PERSISTENT -- LIFE func: odb_is_persistent
	 odb_is_persistent(+term)
*/

static long c_odb_is_persistent() {
  declare_var(1);
  
  begin_dbbuiltin(1,0);

  ptr_psi_term result = aim->b;
  deref(result);

  ptr_psi_term term = stack_psi_term(4);
  if (args[0]->dbpointer) {
    term->type=true_type;
  }
  else term->type=false_type;
  resid_aim=NULL;
  push_goal(unify, term, result, NULL);

  end_dbbuiltin();
}

int get_fd(ptr_psi_term arg1) {

  deref_ptr(arg1);
  int fd = -1;
  if (equal_types(arg1->type,inputfilesym)) {
    ptr_node n=find(featcmp,STREAM,arg1->attr_list);
    if (n) {
      arg1=(ptr_psi_term)n->data;
      if (arg1->value)
	fd = fileno((FILE *)arg1->value);
    }
  }
  else {
    if (equal_types(arg1->type,odb_file)) {
      if (arg1->value) {
	ObjFile* file = DBCHANNEL(arg1->value);
	fd = file->handle();
      }
    }
    else if (equal_types(arg1->type,stream)) {
      if (arg1->value)
	fd = fileno((FILE *)arg1->value);
    }
  }
  return fd;
}

/********************* C_CAN_LOCK_FILE ***********************/
/*	 C_CAN_LOCK_FILE-- LIFE pred: can_lock_file
	 can_lock_file(int type, +odb_file/stream)
	 type = 0 => is read locked else is write locked
*/
static long c_can_lock_file() {
  declare_var(2);

  types[0] = integer;

  begin_dbbuiltin(2,1);

  ptr_psi_term result = aim->b;
  deref(result);

  int fd = get_fd(args[1]);

  pid_t lk;
  if (fd >= 0) {
    if (INT(0)) 
      lk = is_wlock_file(fd);
    else 
      lk = is_rlock_file(fd);
    ptr_psi_term term = stack_psi_term(4);
    if (lk) {
      term->type=false_type;
    }
    else {
      term->type=true_type;
    }
    resid_aim=NULL;
    push_goal(unify, term, result, NULL);
  }
  else {
    success=FALSE;
    Errorline("bad stream in %P.\n", psiTerm);
  }

  end_dbbuiltin();
}

/********************* C_LOCK_FILE ***********************/
/*	 C_LOCK_FILE-- LIFE pred: lock_file
	 lock_file(+type, +odb_file/stream)
	 type = 0 => read lock else write lock
*/
static long c_lock_file() {
  declare_var(2);

  types[0] = integer;

  begin_dbbuiltin(2,1);

  int fd = get_fd(args[1]);

  int err;
  if (fd >= 0) {
    if (INT(0)) {
      if (err=wlock_file(fd)) {
	success=FALSE;
      }
    }
    else {
      if (err=rlock_file(fd)) {
	success=FALSE;
      }
    }
    //      cerr << "err=" << err << " errno=" << errno << "\n";
  }
  else {
    success=FALSE;
    Errorline("bad stream in %P.\n", psiTerm);
  }

  end_dbbuiltin();

}

/********************* C_UNLOCK_FILE ***********************/
/*	 C_UNLOCK_FILE-- LIFE pred: unlock_file
	 unlock_file(+file_name)
*/
static long c_unlock_file() {
  declare_var(1);

  types[0] = quoted_string;

  begin_dbbuiltin(1,1);

  int fd = open(STRNG(0), O_RDONLY);
  if ( fd == -1 ) {
    if (errno == ENOENT) {
      Errorline("%s: no such file\n", STRNG(0));
      success=FALSE;
    }
    else {
      Errorline("Could not open file %s\n", STRNG(0));
      success=FALSE;
    }
  }
  unlock_file(fd);
  close(fd);

  end_dbbuiltin();
}

/********************* C_ODB_EVICT_TERMS ***********************/
/*	 C_ODB_EVICT_TERMS-- LIFE pred: odb_evict_terms
	 odb_evict_terms(+X, +Y)
	 X == min (% of pages) pages to keep
	 Y == min (% of objects) objects to keep
	 it fails if there no objects or pages were evicted
*/
static long c_odb_evict_terms() {
  declare_var(2);
  types[0] = integer;
  types[1] = integer;

  begin_dbbuiltin(2,2);

  if (object_stream.is_set()) {
    //  cerr << "c code evict objects\n";
    success=object_stream->evict_objects(INT(0), INT(1));
  }
  
  end_dbbuiltin();
}

/********************* C_ODB_STAT ***********************/
/*
  returns LIFE heap memory consumed in bytes by page buffer and object buffer
  @(pn_size=>X, on_size=>Y)
  X== heap memory consumed by page buffer
  Y== heap memory consumed by object buffer
 */
static long c_odb_stat() {
  ptr_psi_term t;
  long success=TRUE;
  long t0, t1,t2;
  
  t=aim->a;
  deref_args(t,set_empty);
  ptr_psi_term result = aim->b;
  deref(result);

  ptr_psi_term term = stack_psi_term(4);

  ptr_psi_term term0 = stack_psi_term(4);
  term0->type=integer;
  term0->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("pn_max_size",&(term->attr_list),term0);

  ptr_psi_term term1 = stack_psi_term(4);
  term1->type=integer;
  term1->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("pn_size",&(term->attr_list),term1);

  ptr_psi_term term2 = stack_psi_term(4);
  term2->type=integer;
  term2->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("on_size",&(term->attr_list),term2);
  
  ptr_psi_term term3 = stack_psi_term(4);
  term3->type=real;
  term3->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("gc_time",&(term->attr_list),term3);
  // gc_time in secs, rest is in microsecs.

  ptr_psi_term term4 = stack_psi_term(4);
  term4->type=integer;
  term4->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("gc_count",&(term->attr_list),term4);

  ptr_psi_term term5 = stack_psi_term(4);
  term5->type=integer;
  term5->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("read_time",&(term->attr_list),term5);

  ptr_psi_term term6 = stack_psi_term(4);
  term6->type=integer;
  term6->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("read_count",&(term->attr_list),term6);

  ptr_psi_term term7 = stack_psi_term(4);
  term7->type=integer;
  term7->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("write_time",&(term->attr_list),term7);

  ptr_psi_term term8 = stack_psi_term(4);
  term8->type=integer;
  term8->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("write_count",&(term->attr_list),term8);

  ptr_psi_term term9 = stack_psi_term(4);
  term9->type=integer;
  term9->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("swizzle_time",&(term->attr_list),term9);

  ptr_psi_term term10 = stack_psi_term(4);
  term10->type=integer;
  term10->value=heap_alloc(sizeof(REAL));
  stack_insert_copystr("swizzle_count",&(term->attr_list),term10);

  if (object_stream.is_set()) {
    t0 =  object_stream->page_buffer()->max_size();
    t1 = object_stream->page_buffer()->size();
    t2 = object_stream->object_buffer()->size();
  }
  else {
    t0=(long)0;
    t1=(long)0;
    t2=(long)0;
  }
  
  *(REAL *)term0->value =(REAL) t0;
  *(REAL *)term1->value =(REAL) t1;
  *(REAL *)term2->value =(REAL) t2;

  *(REAL *)term3->value =(REAL) garbage_time;
  *(REAL *)term4->value =(REAL) times_gc_called;

  *(REAL *)term5->value =(REAL) File:: read_time;
  *(REAL *)term6->value =(REAL) File:: times_read_called;

  *(REAL *)term7->value =(REAL) File:: write_time;
  *(REAL *)term8->value =(REAL) File:: times_write_called;

  *(REAL *)term9->value =(REAL) Object2Psi:: swizzle_time;
  *(REAL *)term10->value =(REAL) Object2Psi:: times_swizzle_called;

  resid_aim=NULL;
  push_goal(unify, term, result, NULL);

  return success;
}

long odb_verbose=FALSE;
/******** C_ODB_VERBOSE
  Toggle the verbose flag & print a message saying whether
  verbose mode is on or off.
*/
static long c_odb_verbose()
{
  ptr_psi_term t;

  t=aim->a;
  deref_args(t,set_empty);
  odb_verbose = !odb_verbose;
  cout << "*** Database Verbose mode is turned "
    << (odb_verbose?"on.\n":"off.\n");
  if (object_stream.is_set()){
    cout << "*** ";
    object_stream->output_heap_info(cout);
    cout << "\n";
  }
  return TRUE;
}

/******** C_ODB_STATISTICS

*/
static long c_odb_statistics()
{
  ptr_psi_term t;

  t=aim->a;
  deref_args(t,set_empty);
  if (object_stream.is_set()){
    //    cout << "*** ";
    object_stream->output_statistics(cout);
    //    cout << "\n";
  }
  return TRUE;
}

static long c_test_list() {
  long success=TRUE;
  
  if (object_stream.is_set()){
    cerr << *object_stream << "\n";
  }
  
  return success;
}

void check_db_definitions() {
  check_definition(&odb_file);
  check_definition(&buffer_group);
  check_definition(&psi_id);
  check_definition(&dbstore_handlersym);

  check_misc_definitions();

#ifdef SYBASE 
  check_rdb_definitions();
#endif

}

void init_db_builtins() {

#ifdef SYBASE 
  init_rdb_builtins();
#endif

  ptr_module curmod = current_module;
  set_current_module(odb_module);

  odb_file = update_symbol(odb_module,"odb_file");
  buffer_group = update_symbol(odb_module, "odb_buffer");
  psi_id = update_symbol(odb_module, "psi_id");
  dbstore_handlersym = update_symbol(odb_module, "dbstore_handler");

  
  new_built_in(odb_module, "odb_init", predicate, c_odb_init);
  new_built_in(odb_module,"odb_store_psi",predicate,c_odb_store_psi);

  new_built_in(odb_module,"odb_open_file",predicate,c_odb_open_file);
  new_built_in(odb_module,"odb_get_file",function,c_odb_get_file);
  new_built_in(odb_module, "odb_set_file_writable",predicate, 
	       c_odb_set_file_writable);

  new_built_in(odb_module,"odb_get_stream",function,c_odb_get_stream);
  new_built_in(odb_module, "odb_flush_stream",predicate, c_odb_flush_stream);
  new_built_in(odb_module,"odb_close_stream",predicate,c_odb_close_stream);

  new_built_in(odb_module, "odb_term2id", function,
	       c_odb_term2id);
  new_built_in(odb_module, "odb_id2term", function,
	       c_odb_id2term);

  new_built_in(odb_module, "odb_fetch_depth", predicate, c_odb_fetch_depth);
  new_built_in(odb_module, "odb_is_persistent", function,
	       c_odb_is_persistent);

  new_built_in(odb_module, "odb_evict_terms", predicate, 
	       c_odb_evict_terms);
  new_built_in(odb_module, "odb_stat", function, 
	       c_odb_stat);
  new_built_in(odb_module, "odb_verbose", predicate, c_odb_verbose);
  new_built_in(odb_module, "odb_statistics", predicate, c_odb_statistics);

  new_built_in(odb_module, "lock_file", predicate, c_lock_file);
  new_built_in(odb_module, "unlock_file", predicate, c_unlock_file);
  new_built_in(odb_module, "can_lock_file", function, c_can_lock_file);

  new_built_in(odb_module, "test_list", predicate, 
	       c_test_list);
  insert_misc_builtins();
  set_current_module(curmod);
}


void new_handler() {
  dberror << "OUT OF VIRTUAL MEMORY";
  throw_error();
}

void call_c_main(long argc, char **argv) {
  set_new_handler(&new_handler);
  c_main(argc, argv);
}

