#ifndef psi_obj_hdr
#define psi_obj_hdr

#include <ostream.h>

class PsiObj;

#include "odb.h"
#include "oid.h"
#include "extern-c++.h"

class ObjectPage;
class ObjFile;

//////////////////////
// CLASS PsiObject //
////////////////////
class PsiObject {	

public:
  enum { START, END };

public:
  PsiObject(ptr_psi_term term, ObjFile*, 
	     int near_hint=END);
  PsiObject(ptr_psi_term term, slot_num snum, ObjectPage& page);
  PsiObject(ptr_psi_term term);
  ~PsiObject(void);

private:
  PsiObject(const PsiObject &obj);	// initializer
  PsiObject& operator=(const PsiObject& obj); // assignment

  void sort2char_1(char*& dp, char* val, 
		   u_1byte typ, unsigned long len);
  void sort2char_2(char*& dp, char* val, 
		   u_1byte typ, unsigned long len);
  void sort_info(ptr_psi_term term, char*& val, 
		   u_1byte& typ , unsigned long& len);

  int psi_tree2objectA(ptr_node n, char* dp, unsigned long& num_of_fields);
  int psi_tree2objectA_2(ptr_node tree, unsigned long&, unsigned long&);  
  void psi_tree2objectB(ptr_node tree, char* dp, 
			unsigned long num_of_fileds, ptr_psi_term root);
  void psi_tree2objectB_2(ptr_node tree, char*& dp, char*& fp, 
			 ptr_psi_term root);

  void psi_term2object(ptr_psi_term term);
  void psi_term2object(ptr_psi_term term, slot_num, ObjectPage&);

  u_1byte check_object_size();

private:
  void create_object(ptr_psi_term term, o_2byte size, const char* data);
  void create_forwarded_object(OId& o_id, slot_num snum,  
			       o_2byte size, const char* data);
  void create_forwarded_object(ObjectPage& page, slot_num snum, 
			       o_2byte size, const char* data);
  void update_object(const ObjectKey& o_key, o_2byte obj_size,
		     const char* data);
       // update_object is used when the modified object size is
       // same as before.
  void sufficient_space(o_2byte size);

private:
  ObjectPage* obj_page;

  unsigned long object_size;
  ptr_psi_term root_sort;
  int top_only;
};

class ObjectBuffer;

#ifdef   CHECK_UNEXPECTED
class psi_loop {

};
#endif

//////////////////////
// CLASS PsiIdTerm //
////////////////////
class PsiIdTerm {

private:
  ptr_psi_term term_d;

public:
  PsiIdTerm(const ObjectKey& o_key);
	  // checks the object_buffer if the o_key is already hashed. 
	  // if so the corresponding psi_id_term is fetched.
	  // if the o_key is not hashed a new psi_id_term is created 
	  // in the LIFE heap and is hashed on o_key.
#ifdef   CHECK_UNEXPECTED
  PsiIdTerm(const ObjectKey& o_key, psi_loop&);
#endif
  PsiIdTerm(ptr_psi_term psi_term);

  operator ptr_psi_term() {deref_ptr(term_d);return term_d;}

  void swizzle();
  void feature_added();
};

//////////////////////
// class Object2Psi //
////////////////////
class Object2Psi {

private:
  ObjFile* file_d;
  ptr_psi_term  term_d;
//  const char* data_d;

public:
  Object2Psi(ObjectKey& o_key):
            file_d(o_key.file()), term_d(object2psi_term(o_key)), 
            root_sort(NULL)
  {
  }
  operator ptr_psi_term();

  friend ostream& operator<<(ostream&, Object2Psi& obj);

private:
  ptr_psi_term object2psi_term(ObjectKey& o_key);
  void object_tree(ptr_node& node, const char* dp);
  void object_tree_2(ptr_node& node, const char* dp, u_2byte num);
  int sort_info(const char*& obj, const char*& sort_value, 
		u_1byte& sort_type, u_2byte& sort_size );
  void char2sort(ptr_psi_term& term, const char* dp, 
		 u_1byte sort_type, u_2byte len);
  void char2psi_id(ptr_psi_term& term, 
		   const char*& data_ptr);

  void output_data(ostream& os);

public:
  // if term_d==NULL, it means its a pointer to another object,
  // coref_pnum, coref_snum are the pointers
  page_num coref_pnum;
  slot_num coref_snum;

private:
  ptr_psi_term root_sort;

public:
  static float swizzle_time;
  static long times_swizzle_called;
};

extern void fetch_db_psi_term(ptr_psi_term& t, int level);

#endif
