#include <iostream.h>

#include "psi_term.h"
#include "oid.h"
#include "exc.h"
#include "psi_obj.h"

extern "C" {
  #include "Source/print.h"
  #include "Source/modules.h"
  ptr_definition new_definition(ptr_keyword);
  void bzero();
}

#define BYTE_COPY(a, b, l) bcopy((const char*)(b), (char*)(a), l)

char* PsiTerm:: psi_to_string(ptr_psi_term t)
{
  if (equal_types(t->type,quoted_string)) {
    if (t->value) {
      return (char *) t->value;
    }
    else {
      return quoted_string->keyword->symbol;
    }
  }
  else {
    return t->type->keyword->symbol;
  }
}

ptr_psi_term PsiTerm:: get_arg (ptr_psi_term psiTerm, char* number)
{
    ptr_node n;
    ptr_psi_term arg1;

    deref_ptr(psiTerm);
    if (n = find (featcmp, number, psiTerm->attr_list))
      if ( arg1 = (ptr_psi_term) n->data ) {
	deref_ptr(arg1);
	return arg1;
      }

    return NULL;
}

GENERIC PsiTerm:: get_value(ptr_psi_term psiTerm,char* attributeName,
			    ptr_psi_term& psi_value)
{
  ptr_node nodeAttr;

  deref_ptr(psiTerm);
  nodeAttr=find(featcmp,attributeName,psiTerm->attr_list);
  if(!nodeAttr) 
    return NULL;
  
  psi_value=(ptr_psi_term)nodeAttr->data;
  deref_ptr(psi_value);
  if(psi_value->value)
    return psi_value->value;

  return NULL;
}

static void n_o_f(ptr_node tree, unsigned long& num) {
  if(tree) {
    if(tree->left)
      n_o_f(tree->left, num);
    if (tree->data) num++;
    if(tree->right)
      n_o_f(tree->right, num);
  }
}

unsigned long PsiTerm:: num_of_features(ptr_node tree) {
  unsigned long num=0;
  n_o_f(tree, num);
  return num;
}

GENERIC PsiTerm:: byte_data_alloc(unsigned long bytes) {
  GENERIC bd = heap_alloc(bytes+sizeof(unsigned long));
  BYTEDATA_SIZE(bd) = bytes;
  bzero((char*)BYTEDATA_DATA(bd), bytes);
  return bd;
}

GENERIC PsiTerm:: byte_data_copy(GENERIC bd) {
  unsigned long bytes = BYTEDATA_SIZE(bd);
  GENERIC bd2 = heap_alloc(bytes+sizeof(unsigned long));
  BYTEDATA_SIZE(bd2) = bytes;
  BYTE_COPY((char*)BYTEDATA_DATA(bd2), (char*)BYTEDATA_DATA(bd), bytes);
  return bd2;
}

ptr_psi_term PsiTerm:: heap_bd_psi_term(ptr_definition sort, 
					unsigned long bytes, 
					long stat)
{
  ptr_psi_term term = heap_psi_term(stat);
  term->type=sort;
  term->value=byte_data_alloc(bytes);
  return term;
}

ptr_psi_term PsiTerm:: heap_bd_psi_term(ptr_psi_term& in,
					ptr_definition sort, 
					unsigned long bytes, 
					long stat)
{
  deref_ptr(in);
  if ((GENERIC)in < heap_pointer) {
    ptr_psi_term term = heap_psi_term(stat);
    term->type=sort;
    term->value=byte_data_alloc(bytes);
    push_psi_ptr_value(in, &(in->coref));
    in->coref=term;
    return term;
  }
  else {
    in->type=sort;
    in->value=byte_data_alloc(bytes);
#ifdef TS
    in->time_stamp=global_time_stamp;
#endif
    in->status=stat;
    return in;
  }
}

ptr_psi_term PsiTerm:: stack_bd_psi_term(ptr_definition sort, 
			       unsigned long bytes, 
			       long stat=4)
{
  ptr_psi_term term = stack_psi_term(stat);
  term->type=sort;
  term->value=byte_data_alloc(bytes);
  return term;
}

ptr_definition PsiTerm:: update_symbol(ptr_module module, 
				       const char* symbol)   
{
  ptr_keyword key;
  ptr_definition result=NULL;
  ptr_int_list opens;
  ptr_module opened;
  ptr_keyword openkey;
  ptr_keyword tempkey;

  /* First clean up the arguments and find out which module to use */

  if(!module) {
    module=extract_module_from_name(symbol);
    if(!module)
      module=current_module;
    symbol=strip_module_name(symbol);
  }

  /* printf("looking up %s#%s\n",module->module_name,symbol); */
  
  /* Now look up 'module#symbol' in the symbol table */
  
  key=hash_lookup(module->symbol_table,symbol);
  
  if(key)
    result=key->definition;
  else
    {
      /* Add 'module#symbol' to the symbol table */
      key=HEAP_ALLOC(struct wl_keyword);
      key->module=module;
      key->symbol=heap_copy_string(symbol);
      key->combined_name=
	heap_copy_string(make_module_token(module,symbol));
      key->c_public=FALSE;
      key->private_feature=FALSE; /*  RM: Mar 11 1993  */
      key->definition=NULL;
      
      hash_insert(module->symbol_table,key->symbol,key);
      
      
      /* Search the open modules of 'module' for 'symbol' */
      opens=module->open_modules;
      openkey=NULL;
      while(opens) {
	opened=(ptr_module)(opens->value);
	if(opened!=module) {
	  
	  tempkey=hash_lookup(opened->symbol_table,symbol);
	  
	  if(tempkey)
	    if(openkey && openkey->c_public && tempkey->c_public) {
	      if(openkey->definition==tempkey->definition) {
		Warningline("benign module name clash: %s and %s\n",
			    openkey->combined_name,
			    tempkey->combined_name);
	      }
	      else {
		Errorline("serious module name clash: \"%s\" and \"%s\"\n",
			  openkey->combined_name,
			  tempkey->combined_name);
		
		result=error_psi_term->type;
	      }
	    }
	    else
	      if(!openkey || !openkey->c_public)
		openkey=tempkey;
	}
	
	opens=opens->next;
      }
      
      if(!result) { /*  RM: Feb  1 1993  */
	
	  if(openkey && openkey->c_public) {
	    /* Found the symbol in an open module */
	    
	    if(!openkey->c_public)
	      Warningline("implicit reference to non-public symbol: %s\n",
			  openkey->combined_name);
	    
	    result=openkey->definition;
	    key->definition=result;
	    
	    /*
	      printf("*** Aliasing %s#%s to %s#%s\n",
	      key->module->module_name,
	      key->symbol,
	      openkey->module->module_name,
	      openkey->symbol);
	      */
	    
	  }
	  else { /* Didn't find it */
	    result=new_definition(key);
	  }
	}
    }
  
  return result;
}

ptr_psi_term PsiTerm:: list_cdr(ptr_psi_term lst)
{
  ptr_psi_term car;
  ptr_psi_term cdr;
  
  
  deref_ptr(lst);
  if(lst->type==alist) {
    get_two_args(lst->attr_list,&car,&cdr);
    if(cdr) {
      deref_ptr(cdr);
      return cdr;
    }
  }
  Errorline("in list_car: psi_term %P is not a list\n",lst);
  return error_psi_term;
}

ptr_psi_term PsiTerm:: list_car(ptr_psi_term lst)
{
  ptr_psi_term car;
  ptr_psi_term cdr;
  
  
  deref_ptr(lst);
  if(lst->type==alist) {
    get_two_args(lst->attr_list,&car,&cdr);
    if(car) {
      deref_ptr(car);
      return car;
    };
  };
  return lst;
}

ptr_psi_term PsiTerm:: StackNil()

{
  ptr_psi_term empty;

  
  empty=stack_psi_term(4);
  empty->type=nil;

  return empty;
}

ptr_psi_term PsiTerm:: StackCons(ptr_psi_term head,ptr_psi_term tail)
{
  ptr_psi_term cons;

  cons=stack_psi_term(4);
  cons->type=alist;
  if(head)
    stack_insert(featcmp,one,&(cons->attr_list),head);
  if(tail)
    stack_insert(featcmp,two,&(cons->attr_list),tail);

  return cons;
}

ptr_psi_term PsiTerm:: StackDisjNil()

{
  ptr_psi_term empty;

  
  empty=stack_psi_term(4);
  empty->type=disj_nil;

  return empty;
}


ptr_psi_term PsiTerm:: StackDisj (ptr_psi_term head, ptr_psi_term tail)
{
  ptr_psi_term t1;

  t1=stack_psi_term(4);
  t1->type=disjunction;
  if(head)
    stack_insert(featcmp,one,&(t1->attr_list),head);
  if(tail)
    stack_insert(featcmp,two,&(t1->attr_list),tail);

  return t1;
}

ptr_psi_term PsiTerm:: psi_term2psi_id(ptr_psi_term pid, ptr_psi_term sid, 
				       ObjFile* f) {

  long success=TRUE;
  page_num pnum;
  slot_num snum;

  deref_ptr(pid);
  pnum=(page_num)*(REAL*)pid->value;

  deref_ptr(sid);
  snum=(slot_num)*(REAL*)sid->value;

  ptr_psi_term res;
  ObjectKey o_key(f,pnum,snum);
  PsiIdTerm id(o_key); // psiidterm call
  res = id;

  return res;

}

ptr_psi_term PsiTerm:: psi_id2psi_term(ptr_psi_term id) {
  deref_ptr(id);
  ptr_psi_term term = stack_psi_term(4);
  ObjectId& object_id = PSIID(id->dbpointer);

  ptr_psi_term t1 = stack_psi_term(4);
  t1->type=integer;
  t1->value=heap_alloc(sizeof(REAL));
  *(REAL *)t1->value =(REAL) object_id.pnum();
  stack_insert_copystr(PID_FLAG,&(term->attr_list),t1);

  t1 = stack_psi_term(4);
  t1->type=integer;
  t1->value=heap_alloc(sizeof(REAL));
  *(REAL *)t1->value =(REAL) object_id.snum();
  stack_insert_copystr(SID_FLAG,&(term->attr_list),t1);
  return term;
}

void PsiTerm:: check_psi_tree(ptr_node tree){

  if(tree) {
    if(tree->left)
      check_psi_tree(tree->left);
    
    if (tree->data) {
      printf("%s: ",tree->key);
      check_psi_term((ptr_psi_term)tree->data);
    }

    if(tree->right)
      check_psi_tree(tree->right);
  }

}

void PsiTerm:: check_psi_term(ptr_psi_term term) {
  deref_ptr(term);
  printf("%s\n",term->type->keyword->symbol);
//  if (term->dbpointer) {
//    printf("%s: %d, ",term->type->keyword->symbol, term->dbpointer);
//    printf("dbpointer is nonzero\n");
//  }
  check_psi_tree(term->attr_list);
}

ostream& operator<<(ostream& os, ptr_psi_term term) {
  display_psi(stdout, term);
  return os;
}
