#include "slist.h"
#include "exc.h"
#include "extern-c++.h"


extern "C" {

void check_bytedata_ptr(GENERIC*);
void check_lonely_ptr(GENERIC*);
void update_ptr(GENERIC*);

}

B_SL_Node:: ~B_SL_Node() {
  //  cerr << "B_SL_Node:: ~B_SL_Node()\n";
  B_SL_Node *p, *n;
  B_SL_Node *node=this;

  p = node->prev3_d;
  n = node->next3_d;
  if ( p && n ) {
    p->next3_d=n;
    n->prev3_d=p;
  }
  else {
    if (p) p->next3_d=NULL;
    if (n) n->prev3_d=NULL;
  }

}

void B_SL_Node:: insert_node(B_SL_Node* p, B_SL_Node* n){ 
  if (p) { 
    p->next3_d=this; 
    prev3_d=p;
  }
  else {
    prev3_d=NULL;
  };
  if (n) {
    n->prev3_d=this; 
    next3_d=n;
  }
  else {
    next3_d=NULL;
  }
}

B_SL_Node* B_SL_Node:: remove_node3(void) {
  B_SL_Node *p, *n;

  B_SL_Node *node=this;
  UNEXPECTED( (this->header_node()),
	     "cannot remove header_node in B_SL_Node::remove_node" )
  p = node->prev3_d;
  n = node->next3_d;
  if ( p && n ) {
    p->next3_d=n;
    n->prev3_d=p;
  }
  else {
    if (p) p->next3_d=NULL;
    if (n) n->prev3_d=NULL;
  }
  node->next3_d=node->prev3_d=NULL;

  return node;

}

void B_SL_Node:: check_data() {
  B_FL_Node:: check_data();
  //  cerr << "SL_Node ";

  if (pass==1) {
    if (VALID_LIFE_ADDRESS(next3_d))
      if (next3_d->header_node()) { 
	check_lonely_ptr((GENERIC*)(&next3_d));
      }      
      else {
	check_bytedata_ptr((GENERIC*)(&next3_d));
      }
    
    if (VALID_LIFE_ADDRESS(prev3_d))
      if (prev3_d->header_node()) { 
	check_lonely_ptr((GENERIC*)(&prev3_d));
      }      
      else {
	check_bytedata_ptr((GENERIC*)(&prev3_d));
      }
  }

  
  if (pass==2) {
    if (VALID_LIFE_ADDRESS(next3_d))
      update_ptr((GENERIC*)(&next3_d));

    if (VALID_LIFE_ADDRESS(prev3_d))
      update_ptr((GENERIC*)(&prev3_d));
  }

  //  cerr << "SL_Node over\n";
}

////////////
// SList //
//////////
SList:: ~SList(){}

void SList:: append(B_SL_Node* node) {

  last_d.prepend3(node);

}

void SList:: prepend(B_SL_Node* node) {
  first_d.append3(node);
}

B_SL_Node* SList:: remove_a(void) {
  B_SL_Node* f = first();

  return f->header_node() ? (B_SL_Node*)NULL : 
                                (B_SL_Node*)f->remove3();

}

B_SL_Node* SList:: remove_e(void) {
  B_SL_Node* l = last();

  return l->header_node() ? (B_SL_Node*)NULL : 
                               (B_SL_Node*)l->remove3();

}


void SList:: delete_a(void) {
  delete remove_a();
}

void SList:: delete_e(void) {
  delete remove_e();
}

void SList:: check_data() {
  //  cerr << "SList ";

  if (pass==1) {
    if (VALID_LIFE_ADDRESS(first_d.next3_d))
      if (first_d.next3_d->header_node()) {
	check_lonely_ptr((GENERIC*)(&(first_d.next3_d)));
      }
      else {
	check_bytedata_ptr((GENERIC*)(&(first_d.next3_d)));
      }
    
    if (VALID_LIFE_ADDRESS(last_d.prev3_d))
      if (last_d.prev3_d->header_node()) {
	check_lonely_ptr((GENERIC*)(&(last_d.prev3_d)));
      }
      else {
	check_bytedata_ptr((GENERIC*)(&(last_d.prev3_d)));
      }
  }

  if (pass==2) {
    if (VALID_LIFE_ADDRESS(first_d.next3_d))
      update_ptr((GENERIC*)(&(first_d.next3_d)));

    if (VALID_LIFE_ADDRESS(last_d.prev3_d))
      update_ptr((GENERIC*)(&(last_d.prev3_d)));
  }

  
  //  cerr << "SList over ";
}

void SList:: output_data(ostream& os)const{
  os << "slist header: "; 
  os << "[";
  OUTPUT(&first_d, first_d.next3_d, first_d.prev3_d);
  B_SL_Node* np = first();
  if (np) {
    np->output_data(os); 
    np = np->next3();
  }
  for ( ; np; np=np->next3() ) {
    os << ", ";
    np->output_data(os);
  };
  OUTPUT(&last_d, last_d.next3_d, last_d.prev3_d);
  os << "]";

}

ostream& operator<<( ostream & os, const SList& list ) {
  list.output_data(os);
  return os;
}


